"""Fixtures for observations."""

from typing import Optional

import pytest

from cg.apps.loqus import LoqusdbAPI
from cg.constants.sequencing import SequencingMethod
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig

from tests.cli.conftest import fixture_base_context
from tests.apps.loqus.conftest import (
    fixture_loqusdb_config,
    fixture_nr_of_loaded_variants,
    fixture_loqusdb_binary_path,
    fixture_loqusdb_config_path,
    fixture_loqusdb_process,
    fixture_loqusdb_api,
)
from tests.models.observations.conftest import (
    fixture_observations_input_files_raw,
    fixture_observations_input_files,
)


class MockLoqusdbAPI(LoqusdbAPI):
    """Mock LoqusdbAPI class."""

    def __init__(self, binary_path: str, config_path: str):
        super().__init__(binary_path, config_path)

    @staticmethod
    def load(*args, **kwargs) -> dict:
        """Mock load method."""
        _ = args
        _ = kwargs
        return dict(variants=15)

    @staticmethod
    def get_case(*args, **kwargs) -> Optional[dict]:
        """Mock get_case method."""
        _ = args
        _ = kwargs
        return {"case_id": "case_id", "_id": "123"}

    @staticmethod
    def get_duplicate(*args, **kwargs) -> Optional[dict]:
        """Mock get_duplicate method."""
        _ = args
        _ = kwargs
        return {"case_id": "case_id"}

    @staticmethod
    def delete_case(*args, **kwargs) -> None:
        """Mock delete_case method."""
        _ = args
        _ = kwargs
        return None


@pytest.fixture(scope="function", name="mock_loqusdb_api")
def fixture_mock_loqusdb_api(filled_file) -> MockLoqusdbAPI:
    """Mock Loqusdb API."""
    return MockLoqusdbAPI(binary_path=filled_file, config_path=filled_file)


@pytest.fixture(scope="function", name="mip_dna_observations_api")
def fixture_mip_dna_observations_api(
    cg_config_object: CGConfig, mock_loqusdb_api: MockLoqusdbAPI
) -> MipDNAObservationsAPI:
    """Rare diseases observations API fixture."""
    mip_dna_observations_api = MipDNAObservationsAPI(cg_config_object, SequencingMethod.WGS)
    mip_dna_observations_api.loqusdb_api = mock_loqusdb_api
    return mip_dna_observations_api



"""Fixtures for backup tests"""

from pathlib import Path

import pytest
from tests.store_helpers import StoreHelpers

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.deliver import DeliverAPI
from cg.store import Store


@pytest.fixture(scope="function", name="deliver_api")
def fixture_deliver_api(
    analysis_store: Store, real_housekeeper_api: HousekeeperAPI, project_dir: Path
) -> DeliverAPI:
    """Fixture for deliver_api

    The fixture will return a delivery api where the store is populated with a case with three individuals.
    The housekeeper database is empty
    """
    _deliver_api = DeliverAPI(
        store=analysis_store,
        hk_api=real_housekeeper_api,
        case_tags=[{"case-tag"}],
        sample_tags=[{"sample-tag"}],
        project_base_path=project_dir,
        delivery_type="balsamic",
    )
    yield _deliver_api


@pytest.fixture(name="delivery_hk_api")
def fixture_delivery_hk_api(
    case_hk_bundle_no_files: dict,
    sample1_cram: Path,
    vcf_file: Path,
    real_housekeeper_api: HousekeeperAPI,
    helpers=StoreHelpers,
) -> HousekeeperAPI:
    """Fixture that returns a housekeeper database with delivery data"""

    case_hk_bundle_no_files["files"] = [
        {"path": str(sample1_cram), "archive": False, "tags": ["cram", "ADM1"]},
        {"path": str(vcf_file), "archive": False, "tags": ["vcf-snv-clinical"]},
    ]
    helpers.ensure_hk_bundle(real_housekeeper_api, bundle_data=case_hk_bundle_no_files)
    return real_housekeeper_api


@pytest.fixture(name="populated_deliver_api")
def fixture_populated_deliver_api(
    analysis_store: Store, delivery_hk_api: HousekeeperAPI, project_dir: Path
) -> DeliverAPI:
    """Return a delivery api where housekeeper is populated with some files"""
    _deliver_api = DeliverAPI(
        store=analysis_store,
        hk_api=delivery_hk_api,
        case_tags=["case-tag"],
        sample_tags=["sample-tag"],
        project_base_path=project_dir,
        delivery_type="balsamic",
    )
    return _deliver_api


@pytest.fixture(name="all_samples_in_inbox")
def fixture_all_samples_in_inbox(tmpdir_factory) -> Path:
    """Fixture that returns a customer inbox path with all samples delivered"""
    inbox = tmpdir_factory.mktemp("inbox")
    Path(inbox, "sample1").mkdir(exist_ok=True, parents=True)
    Path(inbox, "sample2").mkdir(exist_ok=True, parents=True)
    Path(inbox, "case").mkdir(exist_ok=True, parents=True)
    Path(inbox, "sample1", "sample_file.txt").touch(exist_ok=True)
    Path(inbox, "sample2", "sample_file.txt").touch(exist_ok=True)
    Path(inbox, "case", "case_file.txt").touch(exist_ok=True)
    return Path(inbox)


@pytest.fixture(name="samples_missing_in_inbox")
def fixture_samples_missing_in_inbox(tmpdir_factory) -> Path:
    """Fixture that returns a customer inbox path with all samples delivered"""
    inbox = tmpdir_factory.mktemp("inbox")
    Path(inbox, "sample1").mkdir(exist_ok=True, parents=True)
    Path(inbox, "sample2").mkdir(exist_ok=True, parents=True)
    Path(inbox, "sample1", "sample_file.txt").touch(exist_ok=True)
    Path(inbox, "case_with_no_data").mkdir(exist_ok=True, parents=True)
    return Path(inbox)



"""Code that handles CLI commands to upload"""
import logging
import sys
import traceback
from typing import Optional

import click

from cg.cli.upload import vogue
from cg.cli.upload.clinical_delivery import auto_fastq, clinical_delivery
from cg.cli.upload.coverage import coverage
from cg.cli.upload.delivery_report import upload_delivery_report_to_scout
from cg.cli.upload.fohm import fohm
from cg.cli.upload.genotype import genotypes
from cg.cli.upload.gisaid import gisaid
from cg.cli.upload.mutacc import process_solved, processed_solved
from cg.cli.upload.nipt import nipt
from cg.cli.upload.observations import observations, available_observations
from cg.cli.upload.scout import (
    create_scout_load_config,
    scout,
    upload_case_to_scout,
    upload_rna_fusion_report_to_scout,
    upload_rna_to_scout,
    upload_rna_junctions_to_scout,
)
from cg.cli.upload.utils import suggest_cases_to_upload
from cg.cli.upload.validate import validate
from cg.constants import Pipeline
from cg.exc import AnalysisAlreadyUploadedError
from cg.meta.upload.balsamic.balsamic import BalsamicUploadAPI
from cg.meta.upload.mip_dna.mip_dna import MipDNAUploadAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option("-f", "--family", "family_id", help="Upload to all apps")
@click.option(
    "-r",
    "--restart",
    is_flag=True,
    help="Force upload of an analysis that has already been uploaded or marked as started",
)
@click.pass_context
def upload(context: click.Context, family_id: Optional[str], restart: bool):
    """Upload results from analyses"""

    config_object: CGConfig = context.obj
    upload_api = MipDNAUploadAPI(config=config_object)  # default upload API

    LOG.info("----------------- UPLOAD -----------------")

    if context.invoked_subcommand is not None:
        context.obj.meta_apis["upload_api"] = upload_api
    elif family_id:  # Provided case ID without a subcommand: upload everything
        try:
            upload_api.analysis_api.verify_case_id_in_statusdb(case_id=family_id)
            case_obj: models.Family = upload_api.status_db.family(family_id)
            upload_api.verify_analysis_upload(case_obj=case_obj, restart=restart)
        except AnalysisAlreadyUploadedError:
            # Analysis being uploaded or it has been already uploaded
            return

        # Update the upload API based on the data analysis type (MIP-DNA by default)
        if Pipeline.BALSAMIC in case_obj.data_analysis:
            upload_api = BalsamicUploadAPI(config=config_object)

        context.obj.meta_apis["upload_api"] = upload_api
        upload_api.upload(ctx=context, case_obj=case_obj, restart=restart)
        click.echo(click.style(f"{family_id} analysis has been successfully uploaded", fg="green"))
    else:
        suggest_cases_to_upload(status_db=upload_api.status_db)
        raise click.Abort()


@upload.command()
@click.option("--pipeline", type=EnumChoice(Pipeline), help="Limit to specific pipeline")
@click.pass_context
def auto(context: click.Context, pipeline: Pipeline = None):
    """Upload all completed analyses"""

    LOG.info("----------------- AUTO -----------------")

    status_db: Store = context.obj.status_db

    exit_code = 0
    for analysis_obj in status_db.analyses_to_upload(pipeline=pipeline):
        if analysis_obj.family.analyses[0].uploaded_at is not None:
            LOG.warning(
                f"Skipping upload for case {analysis_obj.family.internal_id}. "
                f"It has been already uploaded at {analysis_obj.family.analyses[0].uploaded_at}."
            )
            continue

        case_id = analysis_obj.family.internal_id
        LOG.info("Uploading analysis for case: %s", case_id)
        try:
            context.invoke(upload, family_id=case_id)
        except Exception:
            LOG.error(f"Case {case_id} upload failed")
            LOG.error(traceback.format_exc())
            exit_code = 1

    sys.exit(exit_code)


upload.add_command(auto_fastq)
upload.add_command(clinical_delivery)
upload.add_command(coverage)
upload.add_command(create_scout_load_config)
upload.add_command(fohm)
upload.add_command(genotypes)
upload.add_command(gisaid)
upload.add_command(nipt)
upload.add_command(process_solved)
upload.add_command(processed_solved)
upload.add_command(scout)
upload.add_command(upload_case_to_scout)
upload.add_command(upload_delivery_report_to_scout)
upload.add_command(upload_rna_fusion_report_to_scout)
upload.add_command(upload_rna_junctions_to_scout)
upload.add_command(upload_rna_to_scout)
upload.add_command(observations)
upload.add_command(available_observations)
upload.add_command(validate)
upload.add_command(vogue)



"""API for uploading cancer observations."""

import logging

from cg.store import models
from housekeeper.store.models import Version

from cg.constants.sequencing import SequencingMethod
from cg.meta.observations.observations_api import ObservationsAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import BalsamicObservationsInputFiles

LOG = logging.getLogger(__name__)


class BalsamicObservationsAPI(ObservationsAPI):
    """API to manage Balsamic observations."""

    def __init__(self, config: CGConfig, sequencing_method: SequencingMethod):
        self.sequencing_method: SequencingMethod = sequencing_method
        super().__init__(config)

    def load_observations(
        self, case: models.Family, input_files: BalsamicObservationsInputFiles
    ) -> None:
        """Load observation counts to Loqusdb for a Balsamic case."""
        raise NotImplementedError

    def extract_observations_files_from_hk(
        self, hk_version: Version
    ) -> BalsamicObservationsInputFiles:
        """Extract observations files given a housekeeper version for cancer."""
        raise NotImplementedError

    def delete_case(self, case: models.Family) -> None:
        """Delete cancer case observations from Loqusdb."""
        raise NotImplementedError
